# def find_one(arr, field=None, value=None, fn=None):
# 	if not arr or not len(arr):
# 		return None

# 	if field:
# 		if isinstance(arr[0], dict):
# 			filtered = list(filter(lambda x: x[field] == value, arr))
# 		else:
# 			filtered = list(filter(lambda x: getattr(x, field) == value, arr))
# 	else:
# 		filtered = list(filter(fn, arr))
		
# 	if not len(filtered):
# 		return None
# 	else:
# 		return filtered[-1]
		
# def find_many(arr, field=None, value=None, fn=None):
# 	if not arr or not len(arr):
# 		return []

# 	if field:
# 		if isinstance(arr[0], dict):
# 			filtered = list(filter(lambda x: x[field] == value, arr))
# 		else:
# 			filtered = list(filter(lambda x: getattr(x, field) == value, arr))
# 	else:
# 		filtered = list(filter(fn, arr))
		
# 	return filtered

# find_all = find_many

# def find_closest(arr, target):
#     closest = min(arr, key=lambda x: abs(x - target))
#     return closest


from datetime import datetime
import logging
import logging.handlers
from logging import DEBUG, INFO, WARNING, ERROR
from typing import Optional
from pathlib import Path
import inspect
import os

from .bases import Base
from .systems import mkdir

class Logger():
    # 如果要精确的日志不混淆，例如高频交易，需要吧file_fmt设置为'%Y%m%d_%H%M%S.%f' 毫秒级别
    def __init__(self, path: Optional[Path]=None, name: Optional[str]=None, file_fmt='%Y%m%d'):
        ch = logging.StreamHandler()
        ch.setLevel(logging.INFO) # 指定被处理的信息级别为最低级DEBUG，低于level级别的信息将被忽略
        # 输出到file
        logger_name = name or 'ks_utility'
        logger_path: Path = path or Path.home().joinpath('.ks_logs').joinpath(logger_name)
        logger_path.mkdir(parents=True, exist_ok=True)
        file_name = '%s.log'%(datetime.now().strftime(file_fmt))
        file_path = logger_path.joinpath(file_name)
        fh = logging.handlers.RotatingFileHandler(file_path, mode='a', encoding='utf-8')  # 不拆分日志文件，a指追加模式,w为覆盖模式
        fh.setLevel(logging.DEBUG)
        fh.setFormatter(logging.Formatter("%(asctime)s - %(levelname)s: %(message)s"))

        logger = self.logger = logging.getLogger(logger_name)
        logger.setLevel(logging.DEBUG)

        if not logger.handlers:
            logger.addHandler(ch)
            logger.addHandler(fh)

    def debug(self, msg):
        self.logger.debug(msg)
        
    def info(self, msg):
        self.logger.info(msg)
        
    def warn(self, msg):
        self.logger.warn(msg)
        
    def error(self, msg):
        self.logger.error(msg)

    def log(self, msg='hello_log', tag=[], level=INFO, title=None, name=None, fn_level=1):
        if not title:
            if fn_level == 0:
                caller_frame  = inspect.currentframe()
            elif fn_level == 1:
                caller_frame  = inspect.currentframe().f_back
            elif fn_level == 2:
                caller_frame  = inspect.currentframe().f_back.f_back
            name = name or caller_frame.f_code.co_name

            title = f'[{self.__class__.__name__}.{name}:{caller_frame.f_lineno}]'
        content = f'{title}'

        if tag:
            if not isinstance(tag, list):
                tag = [tag]
            for tag_item in tag:
                if tag_item:
                    content += f'[{tag_item}]'
        content += str(msg)

        self.logger.log(msg=content, level=level)

# root_path = os.path.join(os.path.dirname(__file__), '..')

class LoggerBase(Base):
    def __init__(self, path = None, name=None):
        Base.__init__(self)
        
        self.logger = None # 日志记录器
        self.logger_name = name or self.__class__.__name__
        self.logger_path = path or Path.home().joinpath('.ks_logs').joinpath(self.__class__.__name__) # 项目跟文件夹
        
    def log(self, msg='hello_log', tag=[], level=INFO, title=None, name=None, fn_level=1):
        if not self.logger:
            self.logger = Logger(path=self.logger_path, name=self.logger_name)

        if not title:
            if fn_level == 0:
                caller_frame  = inspect.currentframe()
            elif fn_level == 1:
                caller_frame  = inspect.currentframe().f_back
            elif fn_level == 2:
                caller_frame  = inspect.currentframe().f_back.f_back
            name = name or caller_frame.f_code.co_name

            title = f'[{self.__class__.__name__}.{name}:{caller_frame.f_lineno}]'

        return self.logger.log(msg=msg, tag=tag, level=level, title=title, name=name, fn_level=fn_level)

    def update_logger(self):
        self.init_dir()
        self.logger = Logger(self.module_path)





from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import padding

def encrypt(message, key):
    backend = default_backend()
    cipher = Cipher(algorithms.AES(key), modes.ECB(), backend=backend)

    padder = padding.PKCS7(128).padder()
    padded_data = padder.update(message) + padder.finalize()

    encryptor = cipher.encryptor()
    ciphertext = encryptor.update(padded_data) + encryptor.finalize()

    return ciphertext

def decrypt(ciphertext, key):
    backend = default_backend()
    cipher = Cipher(algorithms.AES(key), modes.ECB(), backend=backend)

    decryptor = cipher.decryptor()
    padded_data = decryptor.update(ciphertext) + decryptor.finalize()

    unpadder = padding.PKCS7(128).unpadder()
    message = unpadder.update(padded_data) + unpadder.finalize()

    return message


try:
    from vika import Vika
except:
    pass
import pandas as pd
import multitasking
from time import sleep
import threading
import traceback

def get(datasheet_id, token):
    if not datasheet_id:
        return pd.DataFrame()

    records_arr = []
    vika = Vika(token)
    datasheet = vika.datasheet(datasheet_id, field_key='name')
    try:
        records = datasheet.records.all()
        for record in records:
            record_obj = record.json()
            record_obj['_id'] = record._id
            records_arr.append(record_obj)
        return pd.DataFrame(records_arr)
    except:
        return pd.DataFrame()
    
def update(datasheet_id, token, kv):
    records_arr = []
    vika = Vika(token)
    datasheet = vika.datasheet(datasheet_id, field_key='name')
    row = datasheet.records.get(ID=kv['ID'])
    row.update(kv)

class VikaManager():
    def __init__(self, token):
        self.vika = Vika(token)
        self.tasks = []
        self.update_data = {}
        self.UPDATE_LEN_LIMIT = 10
        self.loop = True

    def get(self, datasheet_id):
        records_arr = []
        vika = self.vika
        datasheet = vika.datasheet(datasheet_id, field_key='name')
        try:
            records = datasheet.records.all()
            for record in records:
                records_arr.append(record.json())
            return pd.DataFrame(records_arr)
        except:
            return pd.DataFrame()
        
    def update(self, datasheet_id, records, task=False):
        vika = self.vika
        datasheet = vika.datasheet(datasheet_id, field_key='name')
        record_arr = []
        for record in records:
            record_id = record['_id']
            del(record['_id'])
            record_arr.append({'recordId': record_id, 'fields': record})
            if task:
                if not datasheet_id in self.update_data:
                    self.update_data[datasheet_id] = {}
                if not record_id in self.update_data[datasheet_id]:
                    self.update_data[datasheet_id][record_id] = {}
                for k, v in record.items():
                    self.update_data[datasheet_id][record_id][k] = v

        if not task:
            datasheet.update_records(record_arr)

    def task_run(self):
        vika = self.vika
        for datasheet_id, records in self.update_data.items():
            datasheet = vika.datasheet(datasheet_id, field_key='name')
            record_arr = []
            for record_id, record_fields in records.items():
                record_arr.append({'recordId': record_id, 'fields': record_fields})
            index = 0
            while index < len(record_arr):
                datasheet.update_records(record_arr[index:index+self.UPDATE_LEN_LIMIT])
                index += self.UPDATE_LEN_LIMIT
                sleep(1) # 20230707增加 vika好像限制了1秒一次请求
        
        # 复位数据
        self.update_data = {}
            
    @multitasking.task
    def run_task_interval(self, interval=3):
        self.loop = True
        while self.loop:
            # print(f'run,{threading.current_thread().ident}')
            try:
                self.task_run()
            except:
                traceback.print_exc()
            sleep(interval)

    def stop_task_interval(self):
        self.loop = False


class VikaTask():
    def __init__(self, fn):
        self.fn = fn
        self.runed = False

    def run(self):
        self.runed = True
        self.fn()
    

# get('dstWUZMkiVg81oYhEn')


from datetime import datetime
import string
from dateutil.parser import parse
import pytz

DATE_FMT = '%Y-%m-%d'
DATETIME_FMT = DATE_FMT + ' %H:%M:%S.%f' + '%z'

def dt_to_str(time, tpl=DATETIME_FMT):
    if isinstance(time, datetime):
        return time.strftime(tpl)
    elif isinstance(time, str):
        return parse(time).strftime(tpl)
    else:
        return time

#dt_str = dt_to_mq(datetime.now())
#dt_str = dt_to_mq('2022-04-14')

def get_date_str(time):
    return dt_to_str(time, DATE_FMT)

def get_tz(tz_str=''):
    return pytz.timezone(tz_str or 'PRC')

def now(tz=None):
    if not tz:
        tz = get_tz()
    return datetime.now(tz)

def get_utcoffset(dt=None, tz=None):
    if not dt:
        dt = now()
    if tz:
        dt = dt.astimezone(tz)
    zstr = dt.strftime('%z')
    return int(zstr[:3])

# 获取时间对应的取整时间戳
def get_ts_int(dt=None, interval=60):
    if not dt:
        dt = now()
    
    ts = dt.timestamp()

    
    ts_int = int(ts / interval) * interval 

    # 由于时间戳是针对utc时间，所以跨日取整需要调整为相应时区的时间戳，即减去offset的秒数
    if interval >= 86400:
        offset = get_utcoffset(dt)
        ts_int -= offset * 3600
    
    return ts_int



from pyecharts import options as opts
from pyecharts.charts import Kline, Line, Grid
import os

from .datetimes import dt_to_str

def line(bars_df, x, y, title='', bs_fields=[], extra=[], render_path='.'):
    html_path = render_path


    x_data = bars_df[x].tolist()
    
    mark_point_data = []
    if bs_fields:
        for bs_field_index, bs_field in enumerate(bs_fields[1]):
            bs_df = bars_df[bars_df[bs_fields[0]] == bs_field]
            for bs_index, bs in bs_df.iterrows():
                color = 'rgb(255, 0, 0)'
                mark_point_y_value = bs[bs_fields[2][bs_field_index]]
                text = bs[bs_fields[0]]
                amount_label = 'amount_label'
                mark_point_data.append(opts.MarkPointItem(name=amount_label, coord=[bs.date, mark_point_y_value], value=text, \
                symbol='circle', symbol_size=[12, 12], itemstyle_opts=opts.ItemStyleOpts(color=color)))
        

    line = (
        Line(init_opts=opts.InitOpts(
            width="1000px",
            height="800px",
            animation_opts=opts.AnimationOpts(animation=False),
        ))
        .add_xaxis(xaxis_data=x_data)
        .set_global_opts(
            xaxis_opts=opts.AxisOpts(type_="category"),
            yaxis_opts=opts.AxisOpts(is_scale = True),
            legend_opts=opts.LegendOpts(pos_left='50%')
        )
        .set_global_opts(title_opts=opts.TitleOpts(title=title))
    )
    
    for field in y:
        line.add_yaxis(
            series_name=field,
            y_axis=bars_df[field].tolist(),
            is_smooth=True,
            is_hover_animation=False,
            linestyle_opts=opts.LineStyleOpts(width=3, opacity=0.5),
            label_opts=opts.LabelOpts(is_show=False),
            markpoint_opts=opts.MarkPointOpts(data=mark_point_data, label_opts=opts.LabelOpts(formatter='{c}',position='bottom',font_size=14)), 
        )

    if extra:
        # Grid Overlap + Bar
        grid_chart = Grid(
            init_opts=opts.InitOpts(
                width="1800px",
                height="1000px",
                animation_opts=opts.AnimationOpts(animation=False),
            )
        )
        grid_chart.add(
            line,
            grid_opts=opts.GridOpts(pos_left="10%", pos_right="8%", height="30%"),
        )

        height = (90-38)/len(extra)

        for index, ex in enumerate(extra):
            line = Line()
            line.add_xaxis(xaxis_data=x_data)
            line.set_global_opts(
                xaxis_opts=opts.AxisOpts(type_="category"),
                yaxis_opts=opts.AxisOpts(is_scale = True),
                legend_opts=opts.LegendOpts(pos_left=f'{(index+2)*10}%')
            )
            for col in ex:
                line.add_yaxis(
                    series_name=col,
                    y_axis=bars_df[col].tolist(),
                    is_smooth=True,
                    is_hover_animation=False,
                    linestyle_opts=opts.LineStyleOpts(width=3, opacity=0.5),
                    label_opts=opts.LabelOpts(is_show=False),
                )

            top = 14 + (index+1)*height
            if index == 1:
                top = top + 4
            grid_chart.add(
                line,
                grid_opts=opts.GridOpts(
                    pos_left="10%", pos_right=f"10%", pos_top=f"{top}%", height=f"{height}%"
                ),
            )
        grid_chart.render(html_path)
    else:
        line.render(html_path)
    
def kline(bars_df, render_path, extra=[], line=[]):
    html_path = render_path

    symbol = bars_df.iloc[0].symbol

    x_data = list(map(lambda x: dt_to_str(x), bars_df['eob'].tolist()))
    y_data = bars_df[['open', 'close', 'low', 'high']].to_dict('split')['data']
    
    mark_point_data = []
    for index, bar in bars_df[(bars_df.gold == True) | (bars_df.dead == True)].iterrows():
        if bar.gold == True:
            color = 'rgb(255, 0, 0)'
            mark_point_y_value = bar.low
            text = '买'
        else:
            color = 'rgb(0, 255, 0)'
            mark_point_y_value = bar.high
            text = '卖'
        
        amount_label = 'amount_label'
        mark_point_data.append(opts.MarkPointItem(name=amount_label, coord=[dt_to_str(bar.eob), mark_point_y_value], value=text, \
        symbol='circle', symbol_size=[12, 12], itemstyle_opts=opts.ItemStyleOpts(color=color)))
        
    kline = (
        Kline(init_opts=opts.InitOpts(width="1800px", height="800px"))
        .add_xaxis(x_data)
        .add_yaxis(symbol, y_data, 
            markpoint_opts=opts.MarkPointOpts(data=mark_point_data, label_opts=opts.LabelOpts(formatter='{c}',position='bottom',font_size=14)), 
            markline_opts=opts.MarkLineOpts(data=[opts.MarkLineItem(type_="max", value_dim="close")]),
            itemstyle_opts=opts.ItemStyleOpts(
                color="#ec0000",
                color0="#00da3c",
                border_color="#8A0000",
                border_color0="#008F28"
            )
        )
        .set_global_opts(
            legend_opts=opts.LegendOpts(
                is_show=False, pos_bottom=10, pos_left="center"
            ),
            yaxis_opts=opts.AxisOpts(is_scale = True),
            datazoom_opts=[
                opts.DataZoomOpts(
                    is_show=True,
                    type_="inside",
                    # xaxis_index=[0, 1],
                    range_start=0,
                    range_end=100,
                ),
                opts.DataZoomOpts(
                    is_show=True,
                    # xaxis_index=[0, 1],
                    type_="slider",
                    # pos_top="85%",
                    range_start=0,
                    range_end=100,
                ),
            ] if not extra else [
                opts.DataZoomOpts(
                    is_show=True,
                    type_="inside",
                    xaxis_index=[0, 1],
                    range_start=0,
                    range_end=100,
                ),
                opts.DataZoomOpts(
                    is_show=True,
                    xaxis_index=[0, 1],
                    type_="slider",
                    # pos_top="85%",
                    range_start=0,
                    range_end=100,
                ),
            ]
        )
    )

    if line:

        line_chart = Line()
        line_chart.add_xaxis(xaxis_data=x_data)
        for col in line:
            line_chart.add_yaxis(
                series_name=col,
                y_axis=bars_df[col],
                is_smooth=True,
                is_hover_animation=False,
                linestyle_opts=opts.LineStyleOpts(width=1, opacity=0.5),
                label_opts=opts.LabelOpts(is_show=False),
            )
        line_chart.set_global_opts(xaxis_opts=opts.AxisOpts(type_="category"))
        kline = kline.overlap(line_chart)


    #pdb.set_trace()
    if extra:
        # Grid Overlap + Bar
        grid_chart = Grid(
            init_opts=opts.InitOpts(
                width="1800px",
                height="1000px",
                animation_opts=opts.AnimationOpts(animation=False),
            )
        )
        grid_chart.add(
            kline,
            grid_opts=opts.GridOpts(pos_left="10%", pos_right="8%", height="60%"),
        )

        height = (100-38)/len(extra)

        for index, ex in enumerate(extra):
            line = Line()
            line.add_xaxis(xaxis_data=x_data)
            line.set_global_opts(
                xaxis_opts=opts.AxisOpts(type_="category"),
                yaxis_opts=opts.AxisOpts(is_scale = True),
            )
            for col in ex:
                line.add_yaxis(
                    series_name=col,
                    y_axis=bars_df[col].tolist(),
                    is_smooth=True,
                    is_hover_animation=False,
                    linestyle_opts=opts.LineStyleOpts(width=3, opacity=0.5),
                    label_opts=opts.LabelOpts(is_show=False),
                    xaxis_index=1,
                    yaxis_index=1,
                )

            top = 8 + (index+1)*height
            grid_chart.add(
                line,
                grid_opts=opts.GridOpts(
                    pos_left="10%", pos_right=f"10%", pos_top=f"{top}%", height=f"20%"
                ),
            )
        grid_chart.render(html_path)
    else:
        kline.render(html_path)


from sympy import sympify

def cal_exp(exp, values):
    # 将表达式字符串转换为 SymPy 表达式
    expression = sympify(exp)
    expression = expression.subs(values)
    result = expression.evalf()
    return result
# t = cal_exp('a+b+c', {'a': 1, 'b': 2, 'c': 3})


import math
import exchange_calendars as ec
import pytz
import datetime
from decimal import Decimal
import pandas as pd

# def handify(volume, base=100):
#     return math.floor(volume/base)*base

# class Calendar():
#     def __init__(self, exchange='XSHG') -> None:
#         self.calendar = ec.get_calendar(exchange)

#     def net_close(self):
#         return self.calendar.next_open(minute=datetime.now()).astimezone(pytz.timezone('PRC'))

# # 是否20cm
# def is_plus(code):
#     return code.startswith('68') or code.startswith('30')

# def get_upper_limit(code, pre_close):
#     rate = 0.2 if is_plus(code) else 0.1
#     return float(round(Decimal(pre_close) * Decimal((1 + rate)), 2))

# def append_suffix(code):
# 	if code.endswith('.SH') or code.endswith('.SZ'):
# 		return code
		
# 	suffix_map = {
# 		'11': '.SH',
# 		'12': '.SZ',
# 		'15': '.SZ',
# 		'51': '.SH',
# 		'56': '.SH',
# 		'58': '.SH',
# 		'00': '.SZ',
# 		'30': '.SZ',
# 		'43': '.BJ',
# 		'60': '.SH',
# 		'68': '.SH',
# 		'83': '.BJ',
# 		'87': '.BJ'
# 	}
# 	return f'{code}{suffix_map[code[0:2]]}'

# 将日线合成周线
def trans_period(bars, p='w'):
    if not len(bars):
        return pd.DataFrame()
    trans_bars = pd.DataFrame()
    day_bars_by_eob = bars.set_index('eob', drop=False)
    trans_bars['open'] = day_bars_by_eob.resample(p).first().to_period(p)['open']
    trans_bars['high'] = day_bars_by_eob.resample(p).max().to_period(p)['high']
    trans_bars['low'] = day_bars_by_eob.resample(p).min().to_period(p)['low']
    trans_bars['close'] = day_bars_by_eob.resample(p).last().to_period(p)['close']
    if 'volume' in day_bars_by_eob:
        trans_bars['volume'] = day_bars_by_eob.resample(p).sum().to_period(p)['volume']
    if 'money' in day_bars_by_eob:
        trans_bars['money'] = day_bars_by_eob.resample(p).sum().to_period(p)['money']
    trans_bars['eob'] = day_bars_by_eob.resample(p).last().to_period(p)['eob']
    trans_bars['symbol'] = day_bars_by_eob.symbol.iloc[0]

    trans_bars = trans_bars.dropna()
    trans_bars = trans_bars[['symbol', 'eob', 'open', 'high', 'low', 'close']]
    trans_bars.reset_index(drop=True, inplace=True)
    return trans_bars

def round_to(value: float, target: float) -> float:
	"""
	Round price to price tick value.
	"""
	value: Decimal = Decimal(str(value))
	target: Decimal = Decimal(str(target))
	rounded: float = float(int(round(value / target)) * target)
	return rounded



from abc import ABC

def obj_to_string(obj):
	# todo restore
	# return 'test'
	"""
	简单地实现类似对象打印的方法
	:param obj: 对应类的实例
	:return: 实例对象的to_string
	"""
	to_string = str(obj.__class__) + "("
	items = obj.__dict__
	n = 0
	for k in items:
		if k.startswith("_"):
			continue
		to_string = to_string + str(k) + "=" + str(items[k]) + ","
		n += 1
	if n == 0:
		to_string += str(obj.__class__).lower() + ": 'Instantiated objects have no property values'"
	return to_string.rstrip(",") + ")"

def obj_to_dict(obj):
	"""
	"""
	obj_dict = {}
	items = obj.__dict__
	for k in items:
		if k.startswith("_"):
			continue
		obj_dict[k] = items[k]
	return obj_dict

class Base(ABC):
	def __init__(self):
		pass

	def __str__(self):
		return obj_to_string(self)

	def __repr__(self):
		return obj_to_string(self)

	def to_dict(self):
		return obj_to_dict(self)

	def setattr(self, dict):
		for key, value in dict.items():
			setattr(self, key, value)


from decimal import Decimal
from enum import Enum
from datetime import datetime
from typing import Optional
import json

from .datetimes import DATETIME_FMT

# 自定义的序列化函数
def decimal_default(obj):
    if isinstance(obj, Decimal):
        return str(obj)  # 或者 str(obj) 如果你想将Decimal转换为字符串
    elif isinstance(obj, Enum):
        return obj.value
    elif isinstance(obj, datetime):
        return obj.strftime(DATETIME_FMT)
    raise TypeError(f"Object of type {obj.__class__.__name__} is not JSON serializable")

def json_dumps(data, cls: Optional[json.JSONEncoder] = None):
    return json.dumps(
        data,
        indent=4,
        ensure_ascii=False,
        default=decimal_default,
        cls=cls
    )